!=====================================================================
!
!               S p e c f e m 3 D  V e r s i o n  2 . 0
!               ---------------------------------------
!
!          Main authors: Dimitri Komatitsch and Jeroen Tromp
!    Princeton University, USA and University of Pau / CNRS / INRIA
! (c) Princeton University / California Institute of Technology and University of Pau / CNRS / INRIA
!                            April 2011
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along
! with this program; if not, write to the Free Software Foundation, Inc.,
! 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
!
!=====================================================================

!----
!---- assemble the contributions between slices using non-blocking MPI
!----

  subroutine assemble_MPI_vector_ext_mesh(NPROC,NGLOB_AB,array_val, &
                        num_interfaces_ext_mesh,max_nibool_interfaces_ext_mesh, &
                        nibool_interfaces_ext_mesh,ibool_interfaces_ext_mesh, &
                        my_neighbours_ext_mesh)

  implicit none

  include "constants.h"

  integer :: NPROC
  integer :: NGLOB_AB

! array to assemble
  real(kind=CUSTOM_REAL), dimension(NDIM,NGLOB_AB) :: array_val

  integer :: num_interfaces_ext_mesh,max_nibool_interfaces_ext_mesh
  integer, dimension(num_interfaces_ext_mesh) :: nibool_interfaces_ext_mesh,my_neighbours_ext_mesh
  integer, dimension(max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh) :: ibool_interfaces_ext_mesh

  ! local parameters

  ! send/receive temporary buffers
  !real(kind=CUSTOM_REAL), dimension(NDIM,max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh) :: &
  !     buffer_send_vector_ext_mesh,buffer_recv_vector_ext_mesh
  real(kind=CUSTOM_REAL), dimension(:,:,:), allocatable :: buffer_send_vector_ext_mesh
  real(kind=CUSTOM_REAL), dimension(:,:,:), allocatable :: buffer_recv_vector_ext_mesh

  ! requests
  !integer, dimension(num_interfaces_ext_mesh) :: request_send_vector_ext_mesh,request_recv_vector_ext_mesh
  integer, dimension(:), allocatable :: request_send_vector_ext_mesh
  integer, dimension(:), allocatable :: request_recv_vector_ext_mesh

  integer ipoin,iinterface,ier


! here we have to assemble all the contributions between partitions using MPI

! assemble only if more than one partition
  if(NPROC > 1) then

    allocate(buffer_send_vector_ext_mesh(NDIM,max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh),stat=ier)
    if( ier /= 0 ) stop 'error allocating array buffer_send_vector_ext_mesh'
    allocate(buffer_recv_vector_ext_mesh(NDIM,max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh),stat=ier)
    if( ier /= 0 ) stop 'error allocating array buffer_recv_vector_ext_mesh'
    allocate(request_send_vector_ext_mesh(num_interfaces_ext_mesh),stat=ier)
    if( ier /= 0 ) stop 'error allocating array request_send_vector_ext_mesh'
    allocate(request_recv_vector_ext_mesh(num_interfaces_ext_mesh),stat=ier)
    if( ier /= 0 ) stop 'error allocating array request_recv_vector_ext_mesh'

    ! partition border copy into the buffer
    do iinterface = 1, num_interfaces_ext_mesh
      do ipoin = 1, nibool_interfaces_ext_mesh(iinterface)
        buffer_send_vector_ext_mesh(:,ipoin,iinterface) = &
          array_val(:,ibool_interfaces_ext_mesh(ipoin,iinterface))
      enddo
    enddo

    ! send messages
    do iinterface = 1, num_interfaces_ext_mesh
      call issend_cr(buffer_send_vector_ext_mesh(1,1,iinterface), &
           NDIM*nibool_interfaces_ext_mesh(iinterface), &
           my_neighbours_ext_mesh(iinterface), &
           itag, &
           request_send_vector_ext_mesh(iinterface) &
           )
      call irecv_cr(buffer_recv_vector_ext_mesh(1,1,iinterface), &
           NDIM*nibool_interfaces_ext_mesh(iinterface), &
           my_neighbours_ext_mesh(iinterface), &
           itag, &
           request_recv_vector_ext_mesh(iinterface) &
           )
    enddo

    ! wait for communications completion (recv)
    do iinterface = 1, num_interfaces_ext_mesh
      call wait_req(request_recv_vector_ext_mesh(iinterface))
    enddo

    ! adding contributions of neighbours
    do iinterface = 1, num_interfaces_ext_mesh
      do ipoin = 1, nibool_interfaces_ext_mesh(iinterface)
        array_val(:,ibool_interfaces_ext_mesh(ipoin,iinterface)) = &
             array_val(:,ibool_interfaces_ext_mesh(ipoin,iinterface)) &
             + buffer_recv_vector_ext_mesh(:,ipoin,iinterface)
      enddo
    enddo

    ! wait for communications completion (send)
    do iinterface = 1, num_interfaces_ext_mesh
      call wait_req(request_send_vector_ext_mesh(iinterface))
    enddo

    deallocate(buffer_send_vector_ext_mesh)
    deallocate(buffer_recv_vector_ext_mesh)
    deallocate(request_send_vector_ext_mesh)
    deallocate(request_recv_vector_ext_mesh)

  endif

  end subroutine assemble_MPI_vector_ext_mesh

!
!-------------------------------------------------------------------------------------------------
!
  subroutine assemble_MPI_vector_ext_mesh_s(NPROC,NGLOB_AB,array_val, &
            buffer_send_vector_ext_mesh,buffer_recv_vector_ext_mesh, &
            num_interfaces_ext_mesh,max_nibool_interfaces_ext_mesh, &
            nibool_interfaces_ext_mesh,ibool_interfaces_ext_mesh,my_neighbours_ext_mesh, &
            request_send_vector_ext_mesh,request_recv_vector_ext_mesh &
            )

! sends data

  implicit none

  include "constants.h"

  integer :: NPROC
  integer :: NGLOB_AB

! array to assemble
  real(kind=CUSTOM_REAL), dimension(NDIM,NGLOB_AB) :: array_val

  integer :: num_interfaces_ext_mesh,max_nibool_interfaces_ext_mesh

  real(kind=CUSTOM_REAL), dimension(NDIM,max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh) :: &
       buffer_send_vector_ext_mesh,buffer_recv_vector_ext_mesh

  integer, dimension(num_interfaces_ext_mesh) :: nibool_interfaces_ext_mesh,my_neighbours_ext_mesh
  integer, dimension(max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh) :: ibool_interfaces_ext_mesh
  integer, dimension(num_interfaces_ext_mesh) :: request_send_vector_ext_mesh,request_recv_vector_ext_mesh

  integer ipoin,iinterface

! here we have to assemble all the contributions between partitions using MPI

! assemble only if more than one partition
  if(NPROC > 1) then

! partition border copy into the buffer
  do iinterface = 1, num_interfaces_ext_mesh
    do ipoin = 1, nibool_interfaces_ext_mesh(iinterface)
      buffer_send_vector_ext_mesh(:,ipoin,iinterface) = &
        array_val(:,ibool_interfaces_ext_mesh(ipoin,iinterface))
    enddo
  enddo

! send messages
  do iinterface = 1, num_interfaces_ext_mesh
    call issend_cr(buffer_send_vector_ext_mesh(1,1,iinterface), &
         NDIM*nibool_interfaces_ext_mesh(iinterface), &
         my_neighbours_ext_mesh(iinterface), &
         itag, &
         request_send_vector_ext_mesh(iinterface) &
         )
    call irecv_cr(buffer_recv_vector_ext_mesh(1,1,iinterface), &
         NDIM*nibool_interfaces_ext_mesh(iinterface), &
         my_neighbours_ext_mesh(iinterface), &
         itag, &
         request_recv_vector_ext_mesh(iinterface) &
         )
  enddo

  endif

  end subroutine assemble_MPI_vector_ext_mesh_s

!
!-------------------------------------------------------------------------------------------------
!

  subroutine assemble_MPI_vector_ext_mesh_w(NPROC,NGLOB_AB,array_val, &
            buffer_recv_vector_ext_mesh,num_interfaces_ext_mesh,max_nibool_interfaces_ext_mesh, &
            nibool_interfaces_ext_mesh,ibool_interfaces_ext_mesh, &
            request_send_vector_ext_mesh,request_recv_vector_ext_mesh)

! waits for data to receive and assembles

  implicit none

  include "constants.h"

  integer :: NPROC
  integer :: NGLOB_AB

! array to assemble
  real(kind=CUSTOM_REAL), dimension(NDIM,NGLOB_AB) :: array_val

  integer :: num_interfaces_ext_mesh,max_nibool_interfaces_ext_mesh

  real(kind=CUSTOM_REAL), dimension(NDIM,max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh) :: &
       buffer_recv_vector_ext_mesh

  integer, dimension(num_interfaces_ext_mesh) :: nibool_interfaces_ext_mesh
  integer, dimension(max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh) :: ibool_interfaces_ext_mesh
  integer, dimension(num_interfaces_ext_mesh) :: request_send_vector_ext_mesh,request_recv_vector_ext_mesh

  integer ipoin,iinterface

! here we have to assemble all the contributions between partitions using MPI

! assemble only if more than one partition
  if(NPROC > 1) then

! wait for communications completion (recv)
  do iinterface = 1, num_interfaces_ext_mesh
    call wait_req(request_recv_vector_ext_mesh(iinterface))
  enddo

! adding contributions of neighbours
  do iinterface = 1, num_interfaces_ext_mesh
    do ipoin = 1, nibool_interfaces_ext_mesh(iinterface)
      array_val(:,ibool_interfaces_ext_mesh(ipoin,iinterface)) = &
           array_val(:,ibool_interfaces_ext_mesh(ipoin,iinterface)) &
           + buffer_recv_vector_ext_mesh(:,ipoin,iinterface)
    enddo
  enddo

! wait for communications completion (send)
  do iinterface = 1, num_interfaces_ext_mesh
    call wait_req(request_send_vector_ext_mesh(iinterface))
  enddo

  endif

  end subroutine assemble_MPI_vector_ext_mesh_w

  subroutine assemble_MPI_ddt_vector_ext_mesh_s(NPROC,NGLOB_AB,array_val, &
            buffer_recv_ddt_vector_ext_mesh, datatype2d, &
            num_interfaces_ext_mesh,max_nibool_interfaces_ext_mesh, &
            nibool_interfaces_ext_mesh,ibool_interfaces_ext_mesh,my_neighbours_ext_mesh, &
            request_send_vector_ext_mesh,request_recv_vector_ext_mesh)

! sends data

  implicit none

  include 'mpif.h'

  include "constants.h"
  include "precision.h"

  integer :: NPROC
  integer :: NGLOB_AB

! array to assemble
  real(kind=CUSTOM_REAL), dimension(NDIM,NGLOB_AB) :: array_val
  
  integer :: num_interfaces_ext_mesh,max_nibool_interfaces_ext_mesh

  real(kind=CUSTOM_REAL), dimension(NDIM,max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh) :: &
       buffer_recv_ddt_vector_ext_mesh

  integer, dimension(num_interfaces_ext_mesh) :: nibool_interfaces_ext_mesh,my_neighbours_ext_mesh
  integer, dimension(max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh) :: ibool_interfaces_ext_mesh
  integer, dimension(num_interfaces_ext_mesh) :: request_send_vector_ext_mesh,request_recv_vector_ext_mesh

  integer iinterface

  INTEGER :: ierr
  INTEGER, DIMENSION(num_interfaces_ext_mesh), INTENT(IN) :: datatype2d

! here we have to assemble all the contributions between partitions using MPI

! assemble only if more than one partition
  if(NPROC > 1) then

! partition border copy into the buffer
  do iinterface = 1, num_interfaces_ext_mesh
! send messages
    CALL MPI_Isend ( array_val(1,1), &
                     1, datatype2d(iinterface), &
                     my_neighbours_ext_mesh(iinterface), &
                     itag, MPI_COMM_WORLD, &
                     request_send_vector_ext_mesh(iinterface), ierr )
    CALL MPI_Irecv ( buffer_recv_ddt_vector_ext_mesh(1,1, iinterface), &
                     nibool_interfaces_ext_mesh(iinterface)*NDIM, CUSTOM_MPI_TYPE, &
                     my_neighbours_ext_mesh(iinterface), &
                     itag, MPI_COMM_WORLD, &
                     request_recv_vector_ext_mesh(iinterface), ierr )
  enddo

  endif

  end subroutine assemble_MPI_ddt_vector_ext_mesh_s

!
!-------------------------------------------------------------------------------------------------
!

  subroutine assemble_MPI_ddt_vector_ext_mesh_w(NPROC,NGLOB_AB,array_val, &
            buffer_recv_ddt_vector_ext_mesh,num_interfaces_ext_mesh,max_nibool_interfaces_ext_mesh, &
            nibool_interfaces_ext_mesh,ibool_interfaces_ext_mesh, &
            request_send_vector_ext_mesh,request_recv_vector_ext_mesh)

! waits for data to receive and assembles

  implicit none

  include 'mpif.h'

  include "constants.h"

  integer :: NPROC
  integer :: NGLOB_AB

! array to assemble
  real(kind=CUSTOM_REAL), dimension(NDIM,NGLOB_AB) :: array_val

  integer :: num_interfaces_ext_mesh,max_nibool_interfaces_ext_mesh

  real(kind=CUSTOM_REAL), dimension(NDIM,max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh), INTENT(IN) :: &
       buffer_recv_ddt_vector_ext_mesh

  integer, dimension(num_interfaces_ext_mesh) :: nibool_interfaces_ext_mesh
  integer, dimension(max_nibool_interfaces_ext_mesh,num_interfaces_ext_mesh) :: ibool_interfaces_ext_mesh
  integer, dimension(num_interfaces_ext_mesh) :: request_send_vector_ext_mesh,request_recv_vector_ext_mesh

  integer ipoin,iinterface

  INTEGER :: ierr

! here we have to assemble all the contributions between partitions using MPI

! assemble only if more than one partition
  if(NPROC > 1) then

! wait for communications completion (send)
  CALL MPI_Waitall ( num_interfaces_ext_mesh, request_send_vector_ext_mesh, MPI_STATUSES_IGNORE, ierr );

! wait for communications completion (recv), but only for one at a time, so we
! can overlap the copy part (the "compute part") and the communication part
  CALL MPI_Waitany ( num_interfaces_ext_mesh, request_recv_vector_ext_mesh, iinterface, MPI_STATUS_IGNORE, ierr );  

  DO WHILE ( iinterface .NE. MPI_UNDEFINED )
    ! adding contributions of neighbours
    do ipoin = 1, nibool_interfaces_ext_mesh(iinterface)
      array_val(:,ibool_interfaces_ext_mesh(ipoin,iinterface)) = &
           array_val(:,ibool_interfaces_ext_mesh(ipoin,iinterface)) &
           + buffer_recv_ddt_vector_ext_mesh(:,ipoin,iinterface)
    enddo
    CALL MPI_Waitany ( num_interfaces_ext_mesh, request_recv_vector_ext_mesh, iinterface, MPI_STATUS_IGNORE, ierr );
  enddo
 
  endif

  end subroutine assemble_MPI_ddt_vector_ext_mesh_w
